#! /usr/bin/env python
import numpy as np
import config
import rospy
import math
import csv
from observer_ekf import ObserverEKF
from observer_luenberg import ObserverLuenberg
from sensor_class import Sensor
from p2_drone.msg import ObserverEstimate
from geometry_msgs.msg import Vector3

class NodeObserver:
    def __init__(self, case_model, case_parder, case_obs, case_pos, signal_freeze):
        self.heading = Sensor(config.heading_sampT, config.heading_noise_std, np.zeros(3))
        self.heading_rate =Sensor(config.heading__rate_sampT, config.heading__rate_noise_std, \
            np.zeros(3))
        self.APS = Sensor(config.APS_sampT, config.APS_noise_std, config.APS_offset)
        self.pub_obs = rospy.Publisher('observer_pos', ObserverEstimate, queue_size=1)
        self.pub_aps_meas = rospy.Publisher('aps_meas', Vector3, queue_size=1)

        self.case_model = case_model  # 1: D = DL and C=0 in model. 2: D= DL+DNL and C!=0.
        self.case_parder = case_parder  # 1: analytical Jacobian. 2: numerical.
        self.case_obs = case_obs  # 1: EKF, 2: Luenberge
        self.case_pos = case_pos  # 1: acoustic 2: global (NED) position. 3: no APS.
        self.signal_freeze = signal_freeze
        print('case_pos: {}'.format('acoustic' if case_pos == 1 else ('global' if case_pos ==2 else 'no APS')))
        print('case_obs: {}'.format('EKF' if case_obs == 1 else 'Luenberger'))
        print('case_model: {} damping'.format('linear' if case_model == 1 else 'nonlinear'))
        print('case_parder: {} jacobian'.format('analytical' if case_parder == 1 else 'numerical'))
        print('signal_freeze: {}'.format('True' if signal_freeze else 'False'))
        if self.case_obs == 1:
            self.obs = ObserverEKF(self.APS, self.heading, self.heading_rate)
        else:
            self.obs = ObserverLuenberg(self.APS)
    
    def spin(self):
        data_vec = []
        rate = rospy.Rate(config.RATE) 
        while not rospy.is_shutdown():
            if self.case_obs==1:
                self.obs.update(self.case_model, self.case_parder, self.case_pos, self.signal_freeze)
            else:
                self.obs.update(self.case_model, self.case_parder, self.case_pos)

            now = rospy.get_rostime()
            y = [self.obs.y[0], self.obs.y[1], self.obs.y[2], self.obs.y[3]]
            data_vec.append(np.append(np.append(np.append(now.secs+now.nsecs/1e9, self.obs.x_hat),y),self.obs.angle_u2n))
            
            self.publishObserverstates()
            self.publishAPS()

            rate.sleep() 
        
        # Save data to file:
        data_vec = np.array(data_vec)
        with open('../test_data/obs1.csv','w+') as my_csv:
            csvWriter = csv.writer(my_csv,delimiter=',')
            csvWriter.writerows(data_vec)

    def publishObserverstates(self):
        xhat = ObserverEstimate()
        xhat.eta_hat = self.nparray2vec3(self.obs.x_hat[0:3])
        xhat.nu_hat = self.nparray2vec3(self.obs.x_hat[3:6])
        xhat.b_hat= self.nparray2vec3(self.obs.x_hat[6:9])
        self.pub_obs.publish(xhat)
    
    def publishAPS(self):
        pos_vec = Vector3()
        pos_vec.x = self.obs.y[0]
        pos_vec.y = self.obs.y[1]
        self.pub_aps_meas.publish(pos_vec)

    def nparray2vec3(self, nparr):
        temp = Vector3()
        temp.x = nparr[0]
        temp.y = nparr[1]
        temp.z = nparr[2]
        return temp

      
if __name__ == '__main__':
    rospy.init_node('observer_ekf', anonymous=True)
    case_model = 1
    case_parder = 1
    case_obs = 1
    case_pos = 1
    signal_freeze = True
    if case_pos == 3:
        signal_freeze = False
    
    observer = NodeObserver(case_model, case_parder, case_obs, case_pos, signal_freeze) 
    observer.spin()
