#!/usr/bin/env python
import numpy as np 
from numpy.linalg import multi_dot
import math
from model import Model
from getmeas_drone import DroneMeas
from getposition_aps import get_global_position, get_acoustic_position, get_master_orientation
import config

class ObserverEKF:
    def __init__(self, APS, heading, heading_rate):   
        self.sampT_EKF = 1/config.RATE 
        self.model = Model(self.sampT_EKF)
        self.x = np.zeros(9,float) 
        self.y = np.zeros(4,float)
        self.tau = np.zeros(3)
        #EKF params
        self.P = np.eye(9)*1e-3
        self.x_hat = self.x
        self.P_hat = self.P
        self.R = np.diag([APS.noise_std[0]**2, APS.noise_std[1]**2, heading.noise_std[0]**2,\
                 heading_rate.noise_std[0]**2])
                 
        self.Q = np.diag([1e1,1e1,1e1]) 
        self.K = np.zeros((9,4),float)

        self.drone_meas = DroneMeas()
        self.APS_offset = APS.offset
        self.angle_u2n = 0
        self.y_prev_list = []
        self.sub_length = 5
        
    def update(self, case_model, case_parder, case_pos, signal_freeze):
        self.get_measurements(case_pos)
        self.y_prev_list.append(np.copy(self.y))
        if (len(self.y_prev_list) == self.sub_length+1):
            self.y_prev_list = self.y_prev_list[1:len(self.y_prev_list)]

        self.model.update(self.x_hat, self.sampT_EKF, case_model, case_parder) 
        #---------EKF-------------------------
        # Check if Kalman gain is singular:
        temp = multi_dot([self.model.H, self.P, self.model.H.T]) + self.R
        try:
            temp2 = np.linalg.inv(temp)
        except:
            temp2 = np.zeros((4,4))
            print('K set to zero due to non-invertible')
        self.K = multi_dot([self.P, self.model.H.T, temp2]) 

        #Check for signal freeze: 
        if signal_freeze:
            y_prev_arr = np.array(self.y_prev_list)
            for i in range(0,len(self.y)):
                if np.var(y_prev_arr[:,i]) == 0:
                    self.K[:,i] = 0.0

        #predict:
        self.x = self.x_hat + \
            self.sampT_EKF*(multi_dot([self.model.A, self.x_hat]) + multi_dot([self.model.B, self.tau]))
        self.x[2] = self.rad2radpi(self.x[2])
        self.P = multi_dot([self.model.phi, self.P_hat, self.model.phi.T]) \
            + multi_dot([self.model.Gamma, self.Q, self.model.Gamma.T])

        #update:
        y_err = (self.y - multi_dot([self.model.H,self.x]))
        y_err[2] = self.rad2radpi(y_err[2])
        self.x_hat = self.x + multi_dot([self.K,y_err]) 
        self.x_hat[2] = self.rad2radpi(self.x_hat[2])
        self.P_hat = multi_dot([
            (np.eye(9) - multi_dot([self.K, self.model.H])),\
            self.P, \
            (np.eye(9) - multi_dot([self.K, self.model.H])).T]) + \
            multi_dot([self.K, self.R, self.K.T])
        
    def get_measurements(self, case_pos):
        #-------GET MEASUREMENTS------------------------
        # y = [x,y,psi,heading_rate]
        self.y[2] = self.rad2radpi(self.drone_meas.attitude.z*math.pi/180)  # [rad]
        self.y[3] = self.drone_meas.state_estimate.heading_rate*math.pi/180  # [rad/s]
        self.tau = np.array([self.drone_meas.control_force.surge,\
            self.drone_meas.control_force.sway,\
            self.drone_meas.control_force.yaw]).T
        # x and y positiong:     
        url = 'http://192.168.0.1/'
        #User frame:
        if case_pos == 1:
            pos = get_acoustic_position(url)
            xy_u = np.array([pos["x"],pos["y"] ])
            xy_n = self.UserToNED(xy_u, url)
            xy_n = self.transform_CO(xy_n)
            self.y[0] = xy_n[0]
            self.y[1] = xy_n[1]
            
        elif case_pos == 2:
            pos = get_global_position(url)
            self.y[0] = pos["lat"]
            self.y[1] = pos["lon"]
        else:
           self.y[0] = 0
           self.y[1] = 0

    def UserToNED(self, xy_u, url):
        #Rotate 180 deg about x-axis:
        xy = np.matmul(np.diag([1,-1]),xy_u)
        #Rotate about z-axis: 
        self.angle_u2n = (get_master_orientation(url) + 90)*math.pi/180
        xy_n = np.matmul(self.rot_z(self.angle_u2n)[0:2,0:2],xy)
        return xy_n

    def transform_CO(self, xy_n):
        #Transform offset in body frame to NED. #Rotate about z:
        aps_offset = np.matmul(self.rot_z(self.y[2]), self.APS_offset)
        xy = xy_n - aps_offset[0:2]
        return xy
            
    def rot_z(self, angle):
        rot = np.array([[math.cos(angle), -math.sin(angle), 0],\
                        [math.sin(angle), math.cos(angle), 0], \
                        [0,0,1]])
        return rot

    def rad2radpi(self, x):
        # Contain angle between [-pi,pi).
        y = np.arctan2(np.sin(x), np.cos(x))
        if y == math.pi:
            y = -y
        return y